import Migrations from 'structure-migrations'
import pluginsList from '../helpers/plugins'
import Controller from '../helpers/mock-cache-controller'
import MockHTTPServer from '../helpers/mock-http-server'
import OrgTestAPI from 'structure-organizations/test/helpers/test-api'
import AppTestAPI from 'structure-applications/test/helpers/test-api'
import {Cache} from '../../src/index'

const server = new MockHTTPServer()

const orgTestApi = new OrgTestAPI(server)
const appTestApi = new AppTestAPI(server)

async function syncTimeout(ms) {
  return new Promise((resolve) => {
    setTimeout(function() {
      resolve()
    }, ms)
  })
}

/** @test {Cache Middleware} */
describe('cache', function() {

  before(function() {

    this.cache = new Cache()

    this.sandbox = sinon.sandbox.create()
    this.uncachedSpy = this.sandbox.spy(function() {
      return {
        uncached: true
      }
    })
    this.cachedSpy = this.sandbox.spy(function() {
      return {
        cached: true
      }
    })
    this.sandbox.stub(Controller.prototype, 'uncached').callsFake(this.uncachedSpy)
    this.sandbox.stub(Controller.prototype, 'cached').callsFake(this.cachedSpy)

    this.migration = new Migrations({
      db: 'test',
      plugins: pluginsList
    })

    return this.migration.process()

  })

  beforeEach(async function() {

    const orgRes = await orgTestApi.create({
      title: 'Org 1'
    })
    this.orgId = orgRes.body.pkg.id

    const appRes = await appTestApi.create(this.orgId, {
      title: 'App 1'
    })
    this.appId = appRes.body.pkg.id

    const app2Res = await appTestApi.create(this.orgId, {
      title: 'App 2'
    })
    this.app2Id = app2Res.body.pkg.id

  })

  after(function() {
    this.sandbox.restore()
  })

  afterEach(function() {
    this.uncachedSpy.reset()
    this.cachedSpy.reset()
  })

  it('Should not cache uncached routes', async function() {

    const res1 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/uncached`)
      .set('applicationid', this.appId)

    const res2 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/uncached`)
      .set('applicationid', this.appId)

    expect(this.uncachedSpy.callCount).to.equal(2)

    expect(res1.body.pkg).to.deep.equal({
      uncached: true
    })
    expect(res2.body.pkg).to.deep.equal({
      uncached: true
    })

  })

  it('Should cache cached routes', async function() {

    const res1 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.appId)

    const res2 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.appId)

    expect(this.cachedSpy.callCount).to.equal(1)

    expect(res1.body.pkg).to.deep.equal({
      cached: true
    })
    expect(res2.body.pkg).to.deep.equal({
      cached: true
    })

  })

  it('Should expire cache after TTL seconds', async function() {

    this.timeout(10000)

    const res1 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.appId)

    const res2 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.appId)

    // Wait at least 5 seconds as per cache TTL definition
    await syncTimeout(6000)

    const res3 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.appId)

    // res1 should have expired and ignored
    expect(this.cachedSpy.callCount).to.equal(2)

    expect(res1.body.pkg).to.deep.equal({
      cached: true
    })
    expect(res2.body.pkg).to.deep.equal({
      cached: true
    })
    expect(res3.body.pkg).to.deep.equal({
      cached: true
    })

  })

  it('Should scope cache by application', async function() {

    const res1 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.appId)

    const res2 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.app2Id)

    // res1 should cache should not be used
    expect(this.cachedSpy.callCount).to.equal(2)

    expect(res1.body.pkg).to.deep.equal({
      cached: true
    })
    expect(res2.body.pkg).to.deep.equal({
      cached: true
    })

  })

  it('Should invalidate all cached routes for an application', async function() {

    const res1 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.appId)

    const res2 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.app2Id)

    await this.cache.invalidateApplication(this.appId)

    const res3 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.appId)

    const res4 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.app2Id)

    // res1 should have been invalidated but not res2
    expect(this.cachedSpy.callCount).to.equal(3)

    expect(res1.body.pkg).to.deep.equal({
      cached: true
    })
    expect(res2.body.pkg).to.deep.equal({
      cached: true
    })
    expect(res3.body.pkg).to.deep.equal({
      cached: true
    })
    expect(res4.body.pkg).to.deep.equal({
      cached: true
    })

  })

  it('Should skip cache if SKIP_CACHE env variable set', async function() {

    process.env.SKIP_CACHE = true

    const res1 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.appId)

    const res2 = await server
      .get(`/api/${process.env.API_VERSION}/mock-cache-plugin/cached`)
      .set('applicationid', this.appId)

    expect(this.cachedSpy.callCount).to.equal(2)

    process.env.SKIP_CACHE = false

    expect(res1.body.pkg).to.deep.equal({
      cached: true
    })
    expect(res2.body.pkg).to.deep.equal({
      cached: true
    })

  })

})
